package mdp;

import java.util.ArrayList;

public class Bisimulation {
	public MDP mdp;
	
	public ArrayList<ArrayList<Integer>> partitions;

	public Bisimulation(MDP mdp, ArrayList<ArrayList<Integer>> partitions) {
		this.mdp = mdp;
		this.partitions = partitions;
	}

	public Bisimulation(MDP mdp) {
		ArrayList<Integer> p = new ArrayList<Integer>(mdp.stateNo);
		for(int i = 0; i < mdp.stateNo; i++)
			p.add(i);
		
		this.mdp = mdp;
		this.partitions = new ArrayList<ArrayList<Integer>>();
		this.partitions.add(p);
	}
	
	// computes the bisimulation classes. the result is stored in partitions
	//
	public void computeClasses() {
		int oldNo;
		int count = 0;
		do {
			count++;
			oldNo = partitions.size();
			ArrayList<ArrayList<Integer>> newPartitions = new ArrayList<ArrayList<Integer>>();
			
			for(ArrayList<Integer> partition : partitions)
				splitPartition(partition, newPartitions);
			partitions = newPartitions;
			//System.out.println("new part: " + partitions);

		} while(oldNo != partitions.size());
		
		System.out.println(
			"compute bisimulation classes: " + count + " iterations, " +
			partitions.size() + " classes");
	}
	
	// checks if 2 states are bisimilar. computeClass() must be ran before
	//
	public boolean isBisimilar(int state1, int state2) {
		for(ArrayList<Integer> partition : partitions)
			if(partition.contains(state1))
				return partition.contains(state2);
		throw new Error("state " + state1 + " does not belong to any partition");
	}
	
	// split the given partition and adds its pieces to newPartitions
	//
	public void splitPartition(ArrayList<Integer> partition, ArrayList<ArrayList<Integer>> newPartitions) {
		// we have to create a Tba tree for each action. However, many actions
		// won't have transitions for any of the states in the partition. so instead of
		// looping over all actions, we look at the transitions of each state and use
		// only the actions that we find there. however this way we find each action many times,
		// so we flag the visited ones
		boolean[] doneActions = new boolean[mdp.actionNo];
		
		// each element of stateGroups is a set of states of the partition that have the same
		// behaviour for some action. They correspond to the enumeration
		// (\alpha_i, L_i) of page 9 in Baier's paper, but the \alpha's are not used
		// so we don't keep them.
		ArrayList<ArrayList<Integer>> stateGroups = new ArrayList<ArrayList<Integer>>();
		
		// create Tba trees and extract the stateGroups from each of them
		for(int state : partition)
			for(Transition t : mdp.trans[state]) {
				if(doneActions[t.action]) continue;
				doneActions[t.action] = true;
				//System.out.println("computing Tba for action " + t.action);
								
				TbaNode tba = TbaNode.createFromPartition(this, t.action, partition);
				tba.extractStateGroups(stateGroups);
				//System.out.println("stateGroups: " + stateGroups);
			}
		
		// build Tb tree and extract partition from it
		TbNode tb = TbNode.createFromStateGroups(partition, stateGroups);
		tb.extractPartitions(newPartitions);
	}
} 

// Tree L_{B,\alpha}, defined in Baier's paper, p.8
//
class TbaNode {
	ArrayList<TbaNode> children;
	ArrayList<Double> prob;
	int depth;
	ArrayList<Integer> L;		// for leaves
	
	public TbaNode(int depth) {
		this.children = new ArrayList<TbaNode>();
		this.prob = new ArrayList<Double>();
		this.depth = depth;
	}
	
	void extractStateGroups(ArrayList<ArrayList<Integer>> groups) {
		if(L != null)
			groups.add(L);
		else
			for(TbaNode node : children)
				node.extractStateGroups(groups);
	}
	
	// iterative implementation, cause the number of iterations
	// can grow high
	public static void addTransition(TbaNode node, Bisimulation bisim, int state, Transition t) {
		loop: while(node.depth < bisim.partitions.size()) {
			ArrayList<Integer> part = bisim.partitions.get(node.depth);
			double p = t.statesProb(part);

			// check if children with probability p already exists,
			// if so continue with it
			for(int i = 0; i < node.prob.size(); i++)
				if(Math.abs(node.prob.get(i) - p) < 10e-8) {
					node = node.children.get(i);
					continue loop;
				}
			
			// children not found, create it and continue with it
			TbaNode newnode = new TbaNode(node.depth+1);
			node.children.add(newnode);
			node.prob.add(p);
			node = newnode;
		}
		
		if(node.L == null)
			node.L = new ArrayList<Integer>();
		node.L.add(state);
	}
	
	static TbaNode createFromPartition(Bisimulation bisim, int action, ArrayList<Integer> partition) {
		TbaNode root = new TbaNode(0);
		for(int state : partition)
			for(Transition t : bisim.mdp.trans[state])
				if(t.action == action)
					addTransition(root, bisim, state, t);
		return root;
	}
}

//Tree T_B defined in Baier's paper, p.9
//
class TbNode {
	TbNode left, right;
	int depth;
	ArrayList<Integer> C;		// for leaves
	
	public TbNode(int depth) {
		left = right = null;
		this.depth = depth;
	}
	
	// adds the sets on the leaves of the tree to the partition
	void extractPartitions(ArrayList<ArrayList<Integer>> partitions) {
		if(C != null) {
			partitions.add(C);
		} else {
			if(left != null)
				left.extractPartitions(partitions);
			if(right != null)
				right.extractPartitions(partitions);
		}
	}
	
	static void addState(TbNode node, ArrayList<ArrayList<Integer>> stateGroups, int state) {
		while(node.depth < stateGroups.size()) {
			if(stateGroups.get(node.depth).contains(state)) {
				if(node.left == null)
					node.left = new TbNode(node.depth+1);
				node = node.left;
			} else {
				if(node.right == null)
					node.right = new TbNode(node.depth+1);
				node = node.right;
			}
		}
		
		if(node.C == null)
			node.C = new ArrayList<Integer>();
		node.C.add(state);
	}
	
	static TbNode createFromStateGroups(ArrayList<Integer> partition, ArrayList<ArrayList<Integer>> stateGroups) {
		TbNode root = new TbNode(0);
		for(int state : partition)
			addState(root, stateGroups, state);
		return root;
	}
}


